from typing import Callable
from pydantic import BaseModel
from versionhq.tool.model import Tool


def tool(*args):
    """
    Decorator to create a tool from a function.
    """

    def create_tool(tool_name: str) -> Callable:

        def _make_tool(f: Callable) -> Tool:
            if f.__doc__ is None:
                raise ValueError("Function must have a docstring")
            if f.__annotations__ is None:
                raise ValueError("Function must have type annotations")

            class_name = "".join(tool_name.split()).title()
            args_schema = type(
                class_name,
                (BaseModel,),
                {
                    "__annotations__": {
                        k: v for k, v in f.__annotations__.items() if k != "return"
                    },
                },
            )
            return Tool(name=tool_name, func=f, args_schema=args_schema)

        return _make_tool

    if len(args) == 1 and callable(args[0]):
        return create_tool(args[0].__name__)(args[0])

    elif len(args) == 1 and isinstance(args[0], str):
        return create_tool(args[0])

    else:
        raise ValueError("Invalid arguments")



import os
import pytest
from unittest.mock import patch
from typing import Union
from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField, TaskOutput, AgentOutput

DEFAULT_MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_sync_execute_task():
    agent = Agent(
        role="demo agent 1",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=DEFAULT_MODEL_NAME,
        max_tokens=3000,
    )

    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
        context=None,
        callback=None,
    )
    res = task.execute_sync(agent=agent)

    assert isinstance(res, TaskOutput)
    assert res.task_id is task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None

    if hasattr(res.pydantic, "output"):
        assert res.pydantic.output is not None
    else:
        assert hasattr(res.pydantic, "test1")
        if res.pydantic.test1:
            assert type(res.pydantic.test1) == str

        assert hasattr(res.pydantic, "test2")
        if res.pydantic.test2:
            assert type(res.pydantic.test2) == list


def test_async_execute_task():
    agent = Agent(
        role="demo agent 2",
        goal="My amazing goals",
    )

    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
    )

    with patch.object(Agent, "execute_task", return_value="ok") as execute:
        execution = task.execute_async(agent=agent)
        result = execution.result()
        assert result.raw == "ok"
        execute.assert_called_once_with(task=task, context=None)


def test_sync_execute_task_with_task_context():
    """
    Use case = One agent handling multiple tasks sequentially using context set in the main task.
    """

    agent = Agent(
        role="demo agent 3",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=DEFAULT_MODEL_NAME,
        max_tokens=3000,
    )

    sub_task = Task(
        description="analyze the client's business model",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="result", type=str, required=True),
        ]
    )
    main_task = Task(
        description="Define the optimal cohort timeframe in days and target audience.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=str, required=True),
        ],
        context=[sub_task]
    )
    res = main_task.execute_sync(agent=agent)

    assert isinstance(res, TaskOutput)
    assert res.task_id is main_task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None

    if hasattr(res.pydantic, "output"):
        assert res.pydantic.output is not None
    else:
        assert hasattr(res.pydantic, "test1")
        if res.pydantic.test1:
            assert type(res.pydantic.test1) == Union[int, str]

        assert hasattr(res.pydantic, "test2")
        if res.pydantic.test2:
            assert type(res.pydantic.test2) == Union[list, str]

    assert sub_task.output is not None
    assert sub_task.output.json_dict is not None
    assert "result" in main_task.prompt()


def test_sync_execute_task_with_prompt_context():
    """
    Use case:
    - One agent handling multiple tasks sequentially using context set in the main task.
    - On top of that, the agent receives context when they execute the task.
    """

    agent = Agent(
        role="demo agent 4",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=DEFAULT_MODEL_NAME,
        max_tokens=3000,
    )

    sub_task = Task(
        description="analyze the client's business model",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="result", type=str, required=True),
        ]
    )
    main_task = Task(
        description="Define the optimal cohort timeframe in days and target audience.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=str, required=True),
        ],
        context=[sub_task]
    )
    res = main_task.execute_sync(agent=agent, context="plan a Black Friday campaign.")

    assert isinstance(res, TaskOutput)
    assert res.task_id is main_task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None

    if hasattr(res.pydantic, "output"):
        assert res.pydantic.output is not None
    else:
        assert hasattr(res.pydantic, "test1")
        if res.pydantic.test1:
            assert type(res.pydantic.test1) == Union[int, str]

        assert hasattr(res.pydantic, "test2")
        if res.pydantic.test2:
            assert type(res.pydantic.test2) == Union[list, str]

    assert sub_task.output is not None
    assert sub_task.output.json_dict is not None
    assert sub_task.output.pydantic is None

    assert "result" in main_task.prompt()
    assert main_task.prompt_context == "plan a Black Friday campaign."
    assert "plan a Black Friday campaign." in main_task.prompt()


def test_callback():
    """
    See if the callback function is executed well with kwargs.
    """

    def callback_func(item: str = None):
        return f"I am callback with {item}."

    agent = Agent(
        role="demo agent 5",
        goal="My amazing goals",
    )

    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
        ],
        callback=callback_func
    )

    with patch.object(Agent, "execute_task", return_value="ok") as execute:
        execution = task.execute_async(agent=agent, callback_kwargs={"item": "demo for pytest"})
        result = execution.result()
        assert result.raw == "ok"
        execute.assert_called_once_with(task=task, context=None)



def test_delegate():
    agent = Agent(
        role="demo agent 6",
        goal="My amazing goals",
    )

    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
        ],
        allow_delegation=True
    )

    task.execute_sync(agent=agent)

    assert task.output is not None
    assert "delegated_agent" in task.processed_by_agents
    assert task.delegations != 0


# def test_conditional_task():



# tools, CONDITIONAL, token usage



LLM_CONTEXT_WINDOW_SIZES = {
    "gpt-3.5-turbo": 8192,
    "gpt-4": 8192,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4-turbo": 128000,
    "o1-preview": 128000,
    "o1-mini": 128000,
    "deepseek-chat": 128000,
    "gemma2-9b-it": 8192,
    "gemma-7b-it": 8192,
    "llama3-groq-70b-8192-tool-use-preview": 8192,
    "llama3-groq-8b-8192-tool-use-preview": 8192,
    "llama-3.1-70b-versatile": 131072,
    "llama-3.1-8b-instant": 131072,
    "llama-3.2-1b-preview": 8192,
    "llama-3.2-3b-preview": 8192,
    "llama-3.2-11b-text-preview": 8192,
    "llama-3.2-90b-text-preview": 8192,
    "llama3-70b-8192": 8192,
    "llama3-8b-8192": 8192,
    "mixtral-8x7b-32768": 32768,
}


LLM_VARS = {
    "openai": [
        {
            "prompt": "Enter your OPENAI API key (press Enter to skip)",
            "key_name": "OPENAI_API_KEY",
        }
    ],
    "anthropic": [
        {
            "prompt": "Enter your ANTHROPIC API key (press Enter to skip)",
            "key_name": "ANTHROPIC_API_KEY",
        }
    ],
    "gemini": [
        {
            "prompt": "Enter your GEMINI API key (press Enter to skip)",
            "key_name": "GEMINI_API_KEY",
        }
    ],
    "watson": [
        {
            "prompt": "Enter your WATSONX URL (press Enter to skip)",
            "key_name": "WATSONX_URL",
        },
        {
            "prompt": "Enter your WATSONX API Key (press Enter to skip)",
            "key_name": "WATSONX_APIKEY",
        },
        {
            "prompt": "Enter your WATSONX Project Id (press Enter to skip)",
            "key_name": "WATSONX_PROJECT_ID",
        },
    ],
    "ollama": [
        {
            "default": True,
            "API_BASE": "http://localhost:11434",
        }
    ],
    "bedrock": [
        {
            "prompt": "Enter your AWS Access Key ID (press Enter to skip)",
            "key_name": "AWS_ACCESS_KEY_ID",
        },
        {
            "prompt": "Enter your AWS Secret Access Key (press Enter to skip)",
            "key_name": "AWS_SECRET_ACCESS_KEY",
        },
        {
            "prompt": "Enter your AWS Region Name (press Enter to skip)",
            "key_name": "AWS_REGION_NAME",
        },
    ],
    "azure": [
        {
            "prompt": "Enter your Azure deployment name (must start with 'azure/')",
            "key_name": "model",
        },
        {
            "prompt": "Enter your AZURE API key (press Enter to skip)",
            "key_name": "AZURE_API_KEY",
        },
        {
            "prompt": "Enter your AZURE API base URL (press Enter to skip)",
            "key_name": "AZURE_API_BASE",
        },
        {
            "prompt": "Enter your AZURE API version (press Enter to skip)",
            "key_name": "AZURE_API_VERSION",
        },
    ],
    "cerebras": [
        {
            "prompt": "Enter your Cerebras model name (must start with 'cerebras/')",
            "key_name": "model",
        },
        {
            "prompt": "Enter your Cerebras API version (press Enter to skip)",
            "key_name": "CEREBRAS_API_KEY",
        },
    ],
}


PROVIDERS = [
    "openai",
    "anthropic",
    "gemini",
    "ollama",
    "watson",
    "bedrock",
    "azure",
    "cerebras",
    "llama",
]

MODELS = {
    "openai": ["gpt-4", "gpt-4o", "gpt-4o-mini", "o1-mini", "o1-preview"],
    "anthropic": [
        "claude-3-5-sonnet-20240620",
        "claude-3-sonnet-20240229",
        "claude-3-opus-20240229",
        "claude-3-haiku-20240307",
    ],
    "gemini": [
        "gemini/gemini-1.5-flash",
        "gemini/gemini-1.5-pro",
        "gemini/gemini-gemma-2-9b-it",
        "gemini/gemini-gemma-2-27b-it",
    ],
    "ollama": ["ollama/llama3.1", "ollama/mixtral"],
    "watson": [
        "watsonx/meta-llama/llama-3-1-70b-instruct",
        "watsonx/meta-llama/llama-3-1-8b-instruct",
        "watsonx/meta-llama/llama-3-2-11b-vision-instruct",
        "watsonx/meta-llama/llama-3-2-1b-instruct",
        "watsonx/meta-llama/llama-3-2-90b-vision-instruct",
        "watsonx/meta-llama/llama-3-405b-instruct",
        "watsonx/mistral/mistral-large",
        "watsonx/ibm/granite-3-8b-instruct",
    ],
    "bedrock": [
        "bedrock/anthropic.claude-3-5-sonnet-20240620-v1:0",
        "bedrock/anthropic.claude-3-sonnet-20240229-v1:0",
        "bedrock/anthropic.claude-3-haiku-20240307-v1:0",
        "bedrock/anthropic.claude-3-opus-20240229-v1:0",
        "bedrock/anthropic.claude-v2:1",
        "bedrock/anthropic.claude-v2",
        "bedrock/anthropic.claude-instant-v1",
        "bedrock/meta.llama3-1-405b-instruct-v1:0",
        "bedrock/meta.llama3-1-70b-instruct-v1:0",
        "bedrock/meta.llama3-1-8b-instruct-v1:0",
        "bedrock/meta.llama3-70b-instruct-v1:0",
        "bedrock/meta.llama3-8b-instruct-v1:0",
        "bedrock/amazon.titan-text-lite-v1",
        "bedrock/amazon.titan-text-express-v1",
        "bedrock/cohere.command-text-v14",
        "bedrock/ai21.j2-mid-v1",
        "bedrock/ai21.j2-ultra-v1",
        "bedrock/ai21.jamba-instruct-v1:0",
        "bedrock/meta.llama2-13b-chat-v1",
        "bedrock/meta.llama2-70b-chat-v1",
        "bedrock/mistral.mistral-7b-instruct-v0:2",
        "bedrock/mistral.mixtral-8x7b-instruct-v0:1",
    ],
}

JSON_URL = "https://raw.githubusercontent.com/BerriAI/litellm/main/model_prices_and_context_window.json"



import logging
import os
import sys
import threading
import warnings
import litellm
from dotenv import load_dotenv
from litellm import get_supported_openai_params
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Union

from versionhq.llm.llm_vars import LLM_CONTEXT_WINDOW_SIZES
from versionhq.task import TaskOutputFormat
from versionhq.task.model import ResponseField

load_dotenv(override=True)
API_KEY_LITELLM = os.environ.get("API_KEY_LITELLM")
DEFAULT_CONTEXT_WINDOW = int(8192 * 0.75)
os.environ["LITELLM_LOG"] = "DEBUG"


class FilteredStream:
    def __init__(self, original_stream):
        self._original_stream = original_stream
        self._lock = threading.Lock()

    def write(self, s) -> int:
        with self._lock:
            if (
                "Give Feedback / Get Help: https://github.com/BerriAI/litellm/issues/new"
                in s
                or "LiteLLM.Info: If you need to debug this error, use `os.environ['LITELLM_LOG'] = 'DEBUG'`"
                in s
            ):
                return 0
            return self._original_stream.write(s)

    def flush(self):
        with self._lock:
            return self._original_stream.flush()


@contextmanager
def suppress_warnings():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = FilteredStream(old_stdout)
        sys.stderr = FilteredStream(old_stderr)

        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


class LLMResponseSchema:
    """
    Use the response schema for LLM response.
    `field_list` contains the title, value type, bool if required of each field that needs to be returned.
    field_list: [{ title, type, required } ]

    i.e., reponse_schema
    response_type: "array"  *options: "array", "dict"
    propeties: { "recipe_name": { "type": "string" }, },
    required:  ["recipe_name"]
    """

    def __init__(self, response_type: str, field_list: List[ResponseField]):
        self.type = response_type
        self.field_list = field_list

    @property
    def schema(self):
        if len(self.field_list) == 0:
            return

        properties = [
            {
                field.title: {
                    "type": field.type,
                }
            }
            for field in self.field_list
        ]
        required = [field.title for field in self.field_list if field.required == True]
        response_schema = {
            "type": self.type,
            "items": {"type": "object", "properties": {*properties}},
            "required": required,
        }
        return response_schema


class LLM:
    """
    Use LiteLLM to connect with the model of choice.
    (Memo) Response formats will be given at the Task handling.
    """

    def __init__(
        self,
        model: str,
        timeout: Optional[Union[float, int]] = None,
        max_tokens: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        context_window_size: Optional[int] = DEFAULT_CONTEXT_WINDOW,
        callbacks: List[Any] = [],
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        n: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        # response_format: Optional[Dict[str, Any]] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        base_url: Optional[str] = None,
        api_version: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        self.model = model
        self.timeout = timeout
        self.max_tokens = max_tokens
        self.max_completion_tokens = max_completion_tokens
        self.context_window_size = context_window_size
        self.callbacks = callbacks

        self.temperature = temperature
        self.top_p = top_p
        self.n = n
        self.stop = stop
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.logit_bias = logit_bias
        # self.response_format = response_format
        self.seed = seed
        self.logprobs = logprobs
        self.top_logprobs = top_logprobs

        self.base_url = base_url
        self.api_version = api_version
        self.api_key = api_key if api_key else API_KEY_LITELLM

        self.kwargs = kwargs

        litellm.drop_params = True
        litellm.set_verbose=True
        self.set_callbacks(callbacks)

    def call(
        self,
        output_formats: List[TaskOutputFormat],
        field_list: Optional[List[ResponseField]],
        messages: List[Dict[str, str]],
        callbacks: List[Any] = [],
    ) -> str:
        """
        Execute LLM based on Agent's controls.
        """

        with suppress_warnings():
            if callbacks and len(callbacks) > 0:
                self.set_callbacks(callbacks)

            try:
                response_format = None

                #! REFINEME
                if TaskOutputFormat.JSON in output_formats:
                    response_format = LLMResponseSchema(
                        response_type="json_object", field_list=field_list
                    )

                params = {
                    "model": self.model,
                    "messages": messages,
                    "timeout": self.timeout,
                    "temperature": self.temperature,
                    "top_p": self.top_p,
                    "n": self.n,
                    "stop": self.stop,
                    "max_tokens": self.max_tokens or self.max_completion_tokens,
                    "presence_penalty": self.presence_penalty,
                    "frequency_penalty": self.frequency_penalty,
                    "logit_bias": self.logit_bias,
                    # "response_format": response_format,
                    "seed": self.seed,
                    "logprobs": self.logprobs,
                    "top_logprobs": self.top_logprobs,
                    "api_base": self.base_url,
                    "api_version": self.api_version,
                    "api_key": self.api_key,
                    "stream": False,
                    **self.kwargs,
                }
                params = {k: v for k, v in params.items() if v is not None}
                res = litellm.completion(**params)
                return res["choices"][0]["message"]["content"]

            except Exception as e:
                logging.error(f"LiteLLM call failed: {str(e)}")
                return None

    def supports_function_calling(self) -> bool:
        try:
            params = get_supported_openai_params(model=self.model)
            return "response_format" in params
        except Exception as e:
            logging.error(f"Failed to get supported params: {str(e)}")
            return False

    def supports_stop_words(self) -> bool:
        try:
            params = get_supported_openai_params(model=self.model)
            return "stop" in params
        except Exception as e:
            logging.error(f"Failed to get supported params: {str(e)}")
            return False

    def get_context_window_size(self) -> int:
        """
        Only use 75% of the context window size to avoid cutting the message in the middle.
        """
        return (
            int(LLM_CONTEXT_WINDOW_SIZES.get(self.model) * 0.75)
            if hasattr(LLM_CONTEXT_WINDOW_SIZES, self.model)
            else DEFAULT_CONTEXT_WINDOW
        )

    def set_callbacks(self, callbacks: List[Any]):
        callback_types = [type(callback) for callback in callbacks]
        for callback in litellm.success_callback[:]:
            if type(callback) in callback_types:
                litellm.success_callback.remove(callback)

        for callback in litellm._async_success_callback[:]:
            if type(callback) in callback_types:
                litellm._async_success_callback.remove(callback)

        litellm.callbacks = callbacks



import uuid
from abc import ABC
from typing import Any, Dict, List, Callable, Type, Optional, get_args, get_origin
from pydantic import (
    UUID4,
    InstanceOf,
    BaseModel,
    ConfigDict,
    Field,
    create_model,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError

from versionhq.clients.product.model import Product, ProductProvider


class Customer(ABC, BaseModel):
    """
    Store the minimal information on the customer.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="customer's name if any")
    product_list: Optional[List[Product]] = Field(
        default=list, description="store products that the customer is associated with"
    )
    analysis: str = Field(
        default=None, description="store the latest analysis results on the customer"
    )
    on_workflow: bool = Field(
        default=False, description="`True` if they are on some messaging workflows"
    )
    on: Optional[str] = Field(
        default=None, description="destination service for this customer if any"
    )

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError(
                "may_not_set_field", "This field is not to be set by the user.", {}
            )

    def customer_to(self) -> List[ProductProvider]:
        """
        Return list of ProductProvider if the customer has `product_list`
        """

        res = list
        if self.product_list:
            for item in self.product_list:
                if item.provider not in res:
                    res.appned(item.provider)
        return res



import uuid
import warnings
import json
from abc import ABC
from enum import Enum
from dotenv import load_dotenv
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, TYPE_CHECKING, Callable, Optional, Tuple, Union
from pydantic import UUID4, InstanceOf, Json, BaseModel, Field, PrivateAttr, field_validator, model_validator
from pydantic._internal._generate_schema import GenerateSchema
from pydantic_core import PydanticCustomError, core_schema

from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput, ConditionalTask, TaskOutputFormat
from versionhq.task.formatter import create_raw_outputs
from versionhq.team.team_planner import TeamPlanner
from versionhq._utils.logger import Logger
from versionhq._utils.usage_metrics import UsageMetrics


initial_match_type = GenerateSchema.match_type


def match_type(self, obj):
    if getattr(obj, "__name__", None) == "datetime":
        return core_schema.datetime_schema()
    return initial_match_type(self, obj)


GenerateSchema.match_type = match_type

warnings.filterwarnings("ignore", category=SyntaxWarning, module="pysbd")
load_dotenv(override=True)

# agentops = None
# if os.environ.get("AGENTOPS_API_KEY"):
#     try:
#         import agentops  # type: ignore
#     except ImportError:
#         pass


class TaskHandlingProcess(str, Enum):
    """
    Class representing the different processes that can be used to tackle multiple tasks.
    """
    sequential = "sequential"
    hierarchical = "hierarchical"
    consensual = "consensual"


class TeamOutput(BaseModel):
    """
    Store outputs of the tasks handled by the team.
    `json_dict` and `raw` store overall output of tasks that handled by the team,
    while `task_output_list` stores each TaskOutput instance to the tasks handled by the team members.
    Note that `raw` and `json_dict` will be prioritized as TeamOutput to refer over `task_output_list`.
    """

    team_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store the team ID that generate the TeamOutput")
    raw: str = Field(default="", description="raw output of the team lead task handled by the team leader")
    pydantic: Optional[Any] = Field(default=None, description="`raw` converted to the abs. pydantic model")
    json_dict: Union[Dict[str, Any]] = Field(default=None, description="`raw` converted to dictionary")
    task_output_list: list[TaskOutput] = Field(default=list, description="store output of all the tasks that the team has executed")
    token_usage: UsageMetrics = Field(default=dict, description="processed token summary")

    def __str__(self):
        return (str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw)


    def __getitem__(self, key):
        if self.pydantic and hasattr(self.pydantic, key):
            return getattr(self.pydantic, key)
        elif self.json_dict and key in self.json_dict:
            return self.json_dict[key]
        else:
            raise KeyError(f"Key '{key}' not found in the team output.")


    @property
    def json(self) -> Optional[str]:
        if self.tasks_output[-1].output_format != TaskOutputFormat.JSON:
            raise ValueError(
                "No JSON output found in the final task. Please make sure to set the output_json property in the final task in your team."
            )
        return json.dumps(self.json_dict)


    def to_dict(self) -> Dict[str, Any]:
        """
        Convert pydantic / raw output into dict and return the dict.
        When we only have `raw` output, return `{ output: raw }` to avoid an error
        """

        output_dict = {}
        if self.json_dict:
            output_dict.update(self.json_dict)
        elif self.pydantic:
            output_dict.update(self.pydantic.model_dump())
        else:
            output_dict.upate({ "output": self.raw })
        return output_dict


    def return_all_task_outputs(self) -> List[Dict[str, Any]]:
        res = [output.json_dict for output in self.task_output_list]
        return res


class TeamMember(ABC, BaseModel):
    agent: Agent | None = Field(default=None, description="store the agent to be a member")
    is_manager: bool = Field(default=False)
    task: Task | None = Field(default=None)


class Team(BaseModel):
    """
    A collaborative team of agents that handles complex, multiple tasks.
    We define strategies for task executions and overall workflow.
    """

    __hash__ = object.__hash__
    _execution_span: Any = PrivateAttr()
    _logger: Logger = PrivateAttr()
    # _inputs: Optional[Dict[str, Any]] = PrivateAttr(default=None)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None)
    members: List[TeamMember] = Field(default_factory=list, description="store agents' uuids and bool if it is manager")

    # work as a team
    team_tasks: Optional[List[Task]] = Field(default_factory=list, description="optional tasks for the team")
    planning_llm: Optional[Any] = Field(default=None, description="llm to handle the planning of the team tasks (if any)")
    function_calling_llm: Optional[Any] = Field(default=None, description="llm to execute func after all agent execution (if any)")
    prompt_file: str = Field(default="", description="path to the prompt json file to be used by the team.")
    process: TaskHandlingProcess = Field(default=TaskHandlingProcess.sequential)

    # callbacks
    before_kickoff_callbacks: List[Callable[[Optional[Dict[str, Any]]], Optional[Dict[str, Any]]]] = Field(
        default_factory=list,
        description="list of callback functions to be executed before the team kickoff. i.e., adjust inputs"
    )
    after_kickoff_callbacks: List[Callable[[TeamOutput], TeamOutput]] = Field(
        default_factory=list,
        description="list of callback functions to be executed after the team kickoff. i.e., store the result in repo"
    )
    task_callback: Optional[Any] = Field(default=None, description="callback to be executed after each task for all agents execution")
    step_callback: Optional[Any] = Field(default=None, description="callback to be executed after each step for all agents execution")

    verbose: bool = Field(default=True)
    cache: bool = Field(default=True)
    memory: bool = Field(default=False, description="whether the team should use memory to store memories of its execution")
    execution_logs: List[Dict[str, Any]] = Field(default=[], description="list of execution logs for tasks")
    usage_metrics: Optional[UsageMetrics] = Field(default=None, description="usage metrics for all the llm executions")


    def __name__(self) -> str:
        return self.name if self.name is not None else self.id.__str__


    @property
    def key(self) -> str:
        source = [str(member.agent.id.__str__) for member in self.members] + [str(task.id.__str__) for task in self.tasks]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def manager_agent(self) -> Agent:
        manager_agent = [member.agent for member in self.members if member.is_manager == True]
        return manager_agent[0] if len(manager_agent) > 0 else None


    @property
    def manager_task(self) -> Task:
        """
        Aside from the team task, return the task that the `manager_agent` needs to handle.
        The task is set as second priority following to the team tasks.
        """
        task = [member.task for member in self.members if member.is_manager == True]
        return task[0] if len(task) > 0 else None


    @property
    def tasks(self):
        """
        Return all the tasks that the team needs to handle in order of priority:
        1. team tasks,
        2. manager_task,
        3. members' tasks
        """
        sorted_member_tasks = [
            member.task for member in self.members if member.is_manager == True
        ] + [member.task for member in self.members if member.is_manager == False]
        return (
            self.team_tasks + sorted_member_tasks
            if len(self.team_tasks) > 0
            else sorted_member_tasks
        )

    # validators
    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        """Prevent manual setting of the 'id' field by users."""
        if v:
            raise PydanticCustomError("may_not_set_field", "The 'id' field cannot be set by the user.", {})

    # @field_validator("config", mode="before")
    # @classmethod
    # def check_config_type(cls, v: Union[Json, Dict[str, Any]]) -> Union[Json, Dict[str, Any]]:
    #     return json.loads(v) if isinstance(v, Json) else v

    @model_validator(mode="after")
    def check_manager_llm(self):
        """
        Validates that the language model is set when using hierarchical process.
        """

        if self.process == TaskHandlingProcess.hierarchical:
            if self.manager_agent is None:
                raise PydanticCustomError(
                    "missing_manager_llm_or_manager_agent",
                    "Attribute `manager_llm` or `manager_agent` is required when using hierarchical process.",
                    {},
                )

            if (self.manager_agent is not None) and (
                self.members.count(self.manager_agent) > 0
            ):
                raise PydanticCustomError(
                    "manager_agent_in_agents",
                    "Manager agent should not be included in agents list.",
                    {},
                )
        return self


    @model_validator(mode="after")
    def validate_tasks(self):
        """
        Every team member should have a task to handle.
        """
        if self.process == TaskHandlingProcess.sequential:
            for member in self.members:
                if member.task is None:
                    raise PydanticCustomError(
                        "missing_agent_in_task",
                        f"Sequential process error: Agent is missing in the task with the following description: {member.task.description}",
                        {},
                    )
        return self

    @model_validator(mode="after")
    def validate_end_with_at_most_one_async_task(self):
        """
        Validates that the team completes max. one asynchronous task by counting tasks traversed backward
        """

        async_task_count = 0
        for task in reversed(self.tasks):
            if task.async_execution:
                async_task_count += 1
            else:
                break  # stop traversing when a non-async task is found

        if async_task_count > 1:
            raise PydanticCustomError(
                "async_task_count",
                "The team must end with max. one asynchronous task.",
                {},
            )
        return self

    def _get_responsible_agent(self, task: Task) -> Agent:
        res = [member.agent for member in self.members if member.task.id == task.id]
        return None if len(res) == 0 else res[0]

    # setup team planner
    def _handle_team_planning(self):
        team_planner = TeamPlanner(tasks=self.tasks, planner_llm=self.planning_llm)
        result = team_planner._handle_task_planning()

        if result is not None:
            for task in self.tasks:
                task_id = task.id
                task.description += (
                    result[task_id] if hasattr(result, str(task_id)) else result
                )

    # task execution
    def _process_async_tasks(
            self, futures: List[Tuple[Task, Future[TaskOutput], int]], was_replayed: bool = False
        ) -> List[TaskOutput]:
        task_outputs: List[TaskOutput] = []
        for future_task, future, task_index in futures:
            task_output = future.result()
            task_outputs.append(task_output)
            self._process_task_result(future_task, task_output)
            self._store_execution_log(
                future_task, task_output, task_index, was_replayed
            )
        return task_outputs


    def _handle_conditional_task(
        self,
        task: ConditionalTask,
        task_outputs: List[TaskOutput],
        futures: List[Tuple[Task, Future[TaskOutput], int]],
        task_index: int,
        was_replayed: bool,
    ) -> Optional[TaskOutput]:

        if futures:
            task_outputs = self._process_async_tasks(futures, was_replayed)
            futures.clear()

        previous_output = task_outputs[task_index - 1] if task_outputs else None
        if previous_output is not None and not task.should_execute(previous_output):
            self._logger.log(
                "debug",
                f"Skipping conditional task: {task.description}",
                color="yellow",
            )
            skipped_task_output = task.get_skipped_task_output()

            if not was_replayed:
                self._store_execution_log(task, skipped_task_output, task_index)
            return skipped_task_output
        return None


    def _create_team_output(self, task_outputs: List[TaskOutput], lead_task_output: TaskOutput = None) -> TeamOutput:
        """
        Take the output of the first task or the lead task output as the team output `raw` value.
        Note that `tasks` are already sorted by the importance.
        """

        if len(task_outputs) < 1:
            raise ValueError("Something went wrong. Kickoff should return only one task output.")

        final_task_output = lead_task_output if lead_task_output is not None else task_outputs[0]
        # final_string_output = final_task_output.raw
        # self._finish_execution(final_string_output)
        token_usage = self._calculate_usage_metrics()

        return TeamOutput(
            team_id=self.id,
            raw=final_task_output.raw,
            json_dict=final_task_output.json_dict,
            pydantic=final_task_output.pydantic,
            task_output_list=task_outputs,
            token_usage=token_usage,
        )


    def _calculate_usage_metrics(self) -> UsageMetrics:
        """
        Calculate and return the usage metrics that consumed by the team.
        """
        total_usage_metrics = UsageMetrics()

        for member in self.members:
            agent = member.agent
            if hasattr(agent, "_token_process"):
                token_sum = agent._token_process.get_summary()
                total_usage_metrics.add_usage_metrics(token_sum)

        if self.manager_agent and hasattr(self.manager_agent, "_token_process"):
            token_sum = self.manager_agent._token_process.get_summary()
            total_usage_metrics.add_usage_metrics(token_sum)

        self.usage_metrics = total_usage_metrics
        return total_usage_metrics


    def _execute_tasks(self, tasks: List[Task], start_index: Optional[int] = 0, was_replayed: bool = False) -> TeamOutput:
        """
        Executes tasks sequentially and returns the final output in TeamOutput class.
        When we have a manager agent, we will start from executing manager agent's tasks.
        Priority
        1. Team tasks > 2. Manager task > 3. Member tasks (in order of index)
        """

        task_outputs: List[TaskOutput] = []
        lead_task_output: TaskOutput = None
        futures: List[Tuple[Task, Future[TaskOutput], int]] = []
        last_sync_output: Optional[TaskOutput] = None

        for task_index, task in enumerate(tasks):
            if start_index is not None and task_index < start_index:
                if task.output:
                    if task.async_execution:
                        task_outputs.append(task.output)
                    else:
                        task_outputs = [task.output]
                        last_sync_output = task.output
                continue

            responsible_agent = self._get_responsible_agent(task)
            if responsible_agent is None:
                responsible_agent = self.manager_agent if self.manager_agent else self.members[0].agent

            # self._prepare_agent_tools(task)
            # self._log_task_start(task, responsible_agent)

            if isinstance(task, ConditionalTask):
                skipped_task_output = self._handle_conditional_task(task, task_outputs, futures, task_index, was_replayed)
                if skipped_task_output:
                    continue

            if task.async_execution:
                context = create_raw_outputs(tasks=[task, ],task_outputs=([last_sync_output,] if last_sync_output else []))
                future = task.execute_async(agent=responsible_agent, context=context,
                                            # tools=responsible_agent.tools
                                            )
                futures.append((task, future, task_index))
            else:
                if futures:
                    task_outputs = self._process_async_tasks(futures, was_replayed)
                    futures.clear()

                context = create_raw_outputs(tasks=[task,], task_outputs=([ last_sync_output,] if last_sync_output else [] ))
                task_output = task.execute_sync(agent=responsible_agent, context=context,
                                                # tools=responsible_agent.tools
                                                )
                if responsible_agent is self.manager_agent:
                    lead_task_output = task_output

                task_outputs.append(task_output)
                # self._process_task_result(task, task_output)
                # self._store_execution_log(task, task_output, task_index, was_replayed)

        # if futures:
        # task_outputs = self._process_async_tasks(futures, was_replayed)
        return self._create_team_output(task_outputs, lead_task_output)


    def kickoff(self, kwargs_before: Optional[Dict[str, str]] = None, kwargs_after: Optional[Dict[str, Any]] = None) -> TeamOutput:
        """
        Kickoff the team:
        0. Plan the team action if we have `team_tasks` using `planning_llm`.
        1. Address `before_kickoff_callbacks` if any.
        2. Handle team members' tasks in accordance with the `process`.
        3. Address `after_kickoff_callbacks` if any.
        """

        metrics: List[UsageMetrics] = []

        if len(self.team_tasks) > 0 or self.planning_llm is not None:
            self._handle_team_planning()

        if kwargs_before is not None:
            for before_callback in self.before_kickoff_callbacks:
                before_callback(**kwargs_before)

        # self._execution_span = self._telemetry.team_execution_span(self, inputs)
        # self._task_output_handler.reset()
        # self._logging_color = "bold_purple"

        # if inputs is not None:
        #     self._inputs = inputs
        # self._interpolate_inputs(inputs)

        for task in self.tasks:
            if not task.callback:
                task.callback = self.task_callback

        # i18n = I18N(prompt_file=self.prompt_file)

        for member in self.members:
            agent = member.agent
            agent.team = self

            # add the team's common callbacks to each agent.
            if not agent.function_calling_llm:
                agent.function_calling_llm = self.function_calling_llm

            # if agent.allow_code_execution:
            #     agent.tools += agent.get_code_execution_tools()

            if not agent.step_callback:
                agent.step_callback = self.step_callback

        if self.process is None:
            self.process = TaskHandlingProcess.sequential

        result = self._execute_tasks(self.tasks)

        for after_callback in self.after_kickoff_callbacks:
            result = after_callback(result, **kwargs_after)

        metrics += [member.agent._token_process.get_summary() for member in self.members]

        self.usage_metrics = UsageMetrics()
        for metric in metrics:
            self.usage_metrics.add_usage_metrics(metric)

        return result



import os
import pytest
from versionhq.agent.model import Agent
from versionhq.llm.model import LLM

MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_build_agent_with_minimal_input():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
    )

    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent_with_backstory():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management."
    )

    assert agent.role == "analyst"
    assert agent.backstory == "You are competitive analysts who have abundand knowledge in marketing, product management."
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        knowledge="competitor products",
        skillsets=["financial analysis", "product management", ]
    )

    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert agent.knowledge in agent.backstory
    for item in agent.skillsets:
        assert item in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent_with_llm():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        knowledge="competitor products",
        skillsets=["financial analysis", "product management", ],
        llm="gpt-4o"
    )
    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert agent.knowledge in agent.backstory
    for item in agent.skillsets:
        assert item in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-4o"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []

# AGENTS WITH TOOLS



import os
from pydantic import BaseModel

from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField, TaskOutput
from versionhq.team.model import Team, TeamMember, TaskHandlingProcess, TeamOutput
from versionhq._utils.usage_metrics import UsageMetrics

MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")


def test_form_team():
    agent_a = Agent(
        role="agent a",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    agent_b = Agent(
        role="agent b",
        goal="My amazing goals",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    task_1 = Task(
            description="Analyze the client's business model.",
            expected_output_json=True,
            output_field_list=[
                ResponseField(title="test1", type=str, required=True),
                ResponseField(title="test2", type=list, required=True),
            ],
            expected_output_pydantic=True,
            context=[],
            callback=None,
        )

    task_2 = Task(
        description="Define the cohort.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
        context=[],
        callback=None,
    )

    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=True, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )

    assert team.id is not None
    assert team.key is not None
    assert isinstance(team.key, str)
    assert team.manager_agent is not None
    assert team.manager_task.id is task_1.id
    assert len(team.tasks) == 2
    for item in team.tasks:
        assert item.id is task_1.id or item.id is task_2.id


def test_form_team_without_leader():
    agent_a = Agent(
        role="agent a",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    agent_b = Agent(
        role="agent b",
        goal="My amazing goals",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    task_1 = Task(
            description="Analyze the client's business model.",
            expected_output_json=True,
            output_field_list=[
                ResponseField(title="test1", type=str, required=True),
                ResponseField(title="test2", type=list, required=True),
            ],
            expected_output_pydantic=True,
            context=[],
            callback=None,
        )

    task_2 = Task(
        description="Define the cohort.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
        context=[],
        callback=None,
    )

    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )

    assert team.id is not None
    assert team.key is not None
    assert isinstance(team.key, str)
    assert team.manager_agent is None
    assert team.manager_task is None
    assert len(team.tasks) == 2
    for item in team.tasks:
        assert item.id is task_1.id or item.id is task_2.id


def test_kickoff_team_without_leader():
    agent_a = Agent(
        role="Demo Agent A",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    agent_b = Agent(
        role="Demo Agent B-1",
        goal="My amazing goals",
        verbose=True,
        llm=MODEL_NAME,
        max_tokens=3000,
    )

    task_1 = Task(
            description="Analyze the client's business model.",
            expected_output_json=True,
            output_field_list=[
                ResponseField(title="test1", type=str, required=True),
                ResponseField(title="test2", type=list, required=True),
            ],
            expected_output_pydantic=True,
            context=[],
            callback=None,
        )

    task_2 = Task(
        description="Define the cohort.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
        context=[],
        callback=None,
    )

    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )
    res = team.kickoff()
    res_all = res.return_all_task_outputs()

    assert isinstance(res, TeamOutput)
    assert res.team_id is team.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None
    for item in res.task_output_list:
        assert isinstance(item, TaskOutput)

    assert isinstance(res_all, list)
    assert len(res_all) == 2
    for item in res_all:
        assert isinstance(item, dict)
        # if not hasattr(item, "output") and not hasattr(res_all, "output"):
        #     assert "test1" in item
        #     assert "test2" in item
        # else:
        #     assert "output" in item

    assert isinstance(res.token_usage, UsageMetrics)
    assert res.token_usage.total_tokens == 0 # as we dont set token usage on agent



# def test_kickoff_with_team_leader():
#     agent_a = Agent(
#         role="Demo Agent A",
#         goal="My amazing goals",
#         backstory="My amazing backstory",
#         verbose=True,
#         llm=MODEL_NAME,
#         max_tokens=3000,
#     )

#     agent_b1 = Agent(
#         role="Demo Agent B-1",
#         goal="My amazing goals",
#         verbose=True,
#         llm=MODEL_NAME,
#         max_tokens=3000,
#     )

#     task_1 = Task(
#         description="Analyze the client's business model, target audience, and customer information and define the optimal cohort timeframe based on customer lifecycle and product usage patterns.",
#         expected_output_json=True,
#         output_field_list=[
#             ResponseField(title="test1", type=str, required=True),
#             ResponseField(title="test2", type=list, required=True),
#         ],
#         expected_output_pydantic=True,
#         context=[],
#         tools=[],
#         callback=None,
#     )

#     task_2 = Task(
#         description="Amazing task description",
#         expected_output_json=True,
#         expected_output_pydantic=True,
#         output_field_list=[
#             ResponseField(title="test1", type=int, required=True),
#             ResponseField(title="test2", type=list, required=True),
#         ],
#         context=[],
#         tools=[],
#         callback=None,
#     )

#     team_task = Task(
#         description="Amazing team task description",
#         expected_output_json=True,
#         expected_output_pydantic=True,
#         output_field_list=[
#             ResponseField(title="field-1", type=str, required=True),
#         ],
#     )

#     team = Team(
#         members=[
#             TeamMember(agent=agent_a, is_manager=True, task=task_1),
#             TeamMember(agent=agent_b1, is_manager=False, task=task_2),
#         ],
#         team_tasks=[team_task,],
#         process=TaskHandlingProcess.sequential,
#         verbose=True,
#         memory=False,
#         before_kickoff_callbacks=[],  # add any callables
#         after_kickoff_callbacks=[],
#         prompt_file="sample.demo.Prompts.demo.py",
#     )

#     res = team.kickoff()
#     res_all = res.return_all_task_outputs()

#     assert isinstance(res, TeamOutput)
#     assert res.team_id == team.id
#     assert res.raw is not None
#     assert isinstance(res.raw, str)
#     assert res.json_dict is not None
#     assert isinstance(res.json_dict, dict)
#     assert isinstance(res.pydantic, BaseModel)

#     assert isinstance(res_all, list)
#     assert len(res_all) == 2
#     for item in res_all:
#         assert isinstance(item, TaskOutput)
#         assert "test1" in item
#         assert "test2" in item

#     assert isinstance(res.token_usage, UsageMetrics)
#     assert res.token_usage.total_tokens > 0



import os
import pytest
from versionhq.agent.model import Agent
from versionhq.llm.model import LLM
from versionhq.clients.workflow.model import Score, ScoreFormat, MessagingWorkflow, MessagingComponent

MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_store_scores():
    """
    Test if the final result will be calcurated using a random subject
    """

    messaging_component = MessagingComponent(message="demo")
    score_raw = 15
    messaging_component.store_scoring_result("demo", score=score_raw)

    assert messaging_component.score is not None
    assert messaging_component.score.result() is not None



from abc import ABC
from inspect import signature
from typing import Any, Dict, Callable, Type, Optional, get_args, get_origin
from pydantic import (
    InstanceOf,
    BaseModel,
    ConfigDict,
    Field,
    create_model,
    field_validator,
    model_validator,
)

from versionhq._utils.cache_handler import CacheHandler


class Tool(ABC, BaseModel):
    """
    The function that will be executed when the tool is called.
    """

    class ArgsSchema(BaseModel):
        pass

    model_config = ConfigDict()
    name: str = Field(default=None)
    func: Callable = Field(default=None)
    cache_function: Callable = lambda _args=None, _result=None: True
    args_schema: Type[BaseModel] = Field(default_factory=ArgsSchema)
    tool_handler: Optional[Dict[str, Any]] = Field(
        default=None,
        description="store tool_handler to record the usage of this tool. to avoid circular import, set as Dict format",
    )

    @property
    def description(self):
        args_schema = {
            name: { "description": field.description,  "type": Tool._get_arg_annotations(field.annotation) }
            for name, field in self.args_schema.model_fields.items()
        }
        return f"Tool Name: {self.name}\nTool Arguments: {args_schema}\nTool Description: {self.description}"


    @field_validator("args_schema", mode="before")
    @classmethod
    def _default_args_schema(cls, v: Type[BaseModel]) -> Type[BaseModel]:
        if not isinstance(v, cls.ArgsSchema):
            return v

        return type(
            f"{cls.__name__}Schema",
            (BaseModel,),
            {
                "__annotations__": {
                    k: v for k, v in cls._run.__annotations__.items() if k != "return"
                },
            },
        )

    @model_validator(mode="after")
    def set_up_tool_handler_instance(self):
        from versionhq.tool.tool_handler import ToolHandler

        if self.tool_handler:
            ToolHandler(**self.tool_handler)

        return self

    @staticmethod
    def _get_arg_annotations(annotation: type[Any] | None) -> str:
        if annotation is None:
            return "None"

        origin = get_origin(annotation)
        args = get_args(annotation)

        if origin is None:
            return (
                annotation.__name__
                if hasattr(annotation, "__name__")
                else str(annotation)
            )

        if args:
            args_str = ", ".join(Tool._get_arg_annotations(arg) for arg in args)
            return f"{origin.__name__}[{args_str}]"

        return origin.__name__

    def _set_args_schema(self):
        if self.args_schema is None:
            class_name = f"{self.__class__.__name__}Schema"

            self.args_schema = type(
                class_name,
                (BaseModel,),
                {
                    "__annotations__": {
                        k: v
                        for k, v in self._run.__annotations__.items()
                        if k != "return"
                    },
                },
            )

    @classmethod
    def from_composio(
        cls, func: Callable = None, tool_name: str = "Composio tool"
    ) -> "Tool":
        """
        Create a Pydantic BaseModel instance from Composio tools, ensuring the Tool instance has a func to be executed.
        Refer to the `args_schema` from the func signature if any. Else, create an `args_schema`.
        """

        if not func:
            raise ValueError("Params must have a callable 'func' attribute.")

        # args_schema = getattr(tool, "args_schema", None)
        args_fields = {}
        func_signature = signature(func)
        annotations = func_signature.parameters
        for name, param in annotations.items():
            if name != "self":
                param_annotation = (
                    param.annotation if param.annotation != param.empty else Any
                )
                field_info = Field(default=..., description="")
                args_fields[name] = (param_annotation, field_info)

        args_schema = (
            create_model(f"{tool_name}Input", **args_fields)
            if args_fields
            else create_model(f"{tool_name}Input", __base__=BaseModel)
        )

        return cls(name=tool_name, func=func, args_schema=args_schema)

    def run(self, *args, **kwargs) -> Any:
        """
        Use the tool.
        When the tool has a func, execute the func and return any response from the func.
        Else,

        The response will be cascaded to the Task and stored in the TaskOutput.
        """

        print(f"Using the tool: {self.name}")
        result = None

        if self.func is not None:
            result = self.func(
                *args, **kwargs
            )  #! REFINEME - format - json dict, pydantic, raw

        else:
            acceptable_args = self.args_schema.model_json_schema()["properties"].keys()
            arguments = {k: v for k, v in kwargs.items() if k in acceptable_args}
            tool_called = ToolCalled(tool=self, arguments=arguments)

            if self.tool_handler:
                if self.tool_handler.has_called_before(tool_called):
                    self.tool_handler.error = "Agent execution error."

                elif self.tool_handler.cache:
                    result = self.tools_handler.cache.read(
                        tool=tool_called.tool.name, input=tool_called.arguments
                    )
                    # from_cache = result is not None
                    if result is None:
                        result = self.invoke(input=arguments)

            else:
                from versionhq.tool.tool_handler import ToolHandler

                tool_handler = ToolHandler(
                    last_used_tool=tool_called, cache=CacheHandler()
                )
                self.tool_handler = tool_handler
                result = self.invoke(input=arguments)

        return result


class ToolCalled(BaseModel):
    """
    Store the tool called and any kwargs used.
    """

    tool: InstanceOf[Tool] = Field(..., description="store the tool instance to be called.")
    arguments: Optional[Dict[str, Any]] = Field(..., description="kwargs passed to the tool")


class InstructorToolCalled(BaseModel):
    tool: InstanceOf[Tool] = Field(..., description="store the tool instance to be called.")
    arguments: Optional[Dict[str, Any]] = Field(..., description="kwargs passed to the tool")


class CacheTool(BaseModel):
    """
    Default tools to hit the cache.
    """

    name: str = "Hit Cache"
    cache_handler: CacheHandler = Field(default_factory=CacheHandler)

    def hit_cache(self, key):
        split = key.split("tool:")
        tool = split[1].split("|input:")[0].strip()
        tool_input = split[1].split("|input:")[1].strip()
        return self.cache_handler.read(tool, tool_input)



from typing import Any, Dict, Type
from pydantic import BaseModel


def process_config(
    values_to_update: Dict[str, Any], model_class: Type[BaseModel]
) -> Dict[str, Any]:
    """
    Process the config dictionary and update the values accordingly.
    Refer to the Pydantic model class for field validation.
    """

    if hasattr(values_to_update, "config"):
        config = values_to_update.pop("config", {})
    else:
        return values_to_update

    # copy values from config to the model's attributes if the attribute isn't already set.
    for key, value in config.items():
        if key not in model_class.model_fields or values_to_update.get(key) is not None:
            continue

        if isinstance(value, dict) and isinstance(values_to_update.get(key), dict):
            values_to_update[key].update(value)
        else:
            values_to_update[key] = value

    return values_to_update



import os
import warnings

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
print(ROOT_DIR)

warnings.filterwarnings("ignore")
df = pd.read_csv(os.path.join(ROOT_DIR, "/sample_dataset/Online Sales Data.csv"))

df.isna().sum()
df.duplicated().sum()
df.drop(columns="Transaction ID", inplace=True)
df.info()
df.Date.min()
df.Date.max()
df["Date"] = pd.to_datetime(df["Date"])
df["year"] = df["Date"].dt.year
df["month"] = df["Date"].dt.month
df["day"] = df["Date"].dt.day
df["day_name"] = df["Date"].dt.day_name()
df["Product Category"].unique()
df["Product Category"].value_counts()
df["Product Name"].nunique()
df[df["Product Category"] == "Electronics"]["Product Name"]
df[df["Product Category"] == "Home Appliances"]["Product Name"]
df[df["Product Category"] == "Clothing"]["Product Name"]
df[df["Product Category"] == "Books"]["Product Name"]
df[df["Product Category"] == "Beauty Products"]["Product Name"]
df[df["Product Category"] == "Sports"]["Product Name"]
df["Payment Method"].value_counts()
colors = plt.get_cmap("Pastel1").colors
df["Payment Method"].value_counts().plot(
    kind="pie", autopct="%1.1f%%", colors=colors, explode=(0.05, 0, 0)
)
df.Region.unique()
df.Region.value_counts()
colors = plt.get_cmap("Pastel1_r").colors
df["Region"].value_counts().plot(
    kind="pie", autopct="%1.1f%%", colors=colors, explode=(0, 0, 0), startangle=90
)
sales_by_region = df.groupby("Region")["Units Sold"].sum().reset_index()
print(sales_by_region)

sns.barplot(data=sales_by_region, x="Region", y="Units Sold")

total_sales_by_category = (
    df.groupby("Product Category")["Units Sold"].sum().reset_index()
)
print(total_sales_by_category)
plt.figure(figsize=(10, 5))
sns.barplot(data=total_sales_by_category, x="Product Category", y="Units Sold")

total_sales_Payment_Method = (
    df.groupby("Payment Method")["Units Sold"].sum().reset_index()
)
plt.figure(figsize=(7, 5))
sns.barplot(data=total_sales_Payment_Method, x="Payment Method", y="Units Sold")

df["Units Sold"].max()
df[df["Units Sold"] == df["Units Sold"].max()]
data = df[df["Product Category"] == "Electronics"]

Product_Units = (
    data.groupby("Product Name")["Units Sold"]
    .sum()
    .sort_values(ascending=False)
    .reset_index()
)

data[data["Units Sold"].isin(Product_Units["Units Sold"].nlargest(8))]
data = df[df["Product Category"] == "Clothing"]
Product_Units = (
    data.groupby("Product Name")["Units Sold"]
    .sum()
    .sort_values(ascending=False)
    .reset_index()
)
data[data["Units Sold"].isin(Product_Units["Units Sold"].nlargest(3))].sort_values(
    by="Units Sold", ascending=False
)
plt.figure(figsize=(10, 5))
data.groupby("day_name")["Units Sold"].sum()
units_sold_by_day = data.groupby("day_name")["Units Sold"].sum().reset_index()
plt.figure(figsize=(10, 6))
sns.barplot(data=units_sold_by_day, x="day_name", y="Units Sold")

plt.title("Total Units Sold by Day of the Week of Clothing")
plt.xlabel("Day of the Week")
plt.ylabel("Units Sold")
plt.xticks(rotation=45)
plt.tight_layout()
plt.show()

region_counts = data.Region.value_counts()

for region, count in region_counts.items():
    percentage = (count / region_counts.sum()) * 100
    print(f"{region}: {percentage:.1f}%")
Payment_Method = data["Payment Method"].value_counts()

for region, count in Payment_Method.items():
    percentage = (count / Payment_Method.sum()) * 100
    print(f"{region}: {percentage:.1f}%")

data = df[df["Product Category"] == "Books"]
plt.figure(figsize=(10, 5))
data.groupby("day_name")["Units Sold"].count().sort_values(ascending=False)
units_sold_by_day = data.groupby("day_name")["Units Sold"].sum().reset_index()

plt.figure(figsize=(10, 6))
sns.barplot(data=units_sold_by_day, x="day_name", y="Units Sold")

plt.title("Total Units Sold by Day of the Week of Books")
plt.xlabel("Day of the Week")
plt.ylabel("Units Sold")
plt.xticks(rotation=45)
plt.tight_layout()
plt.show()


region_counts = data.Region.value_counts()

for region, count in region_counts.items():
    percentage = (count / region_counts.sum()) * 100
    print(f"{region}: {percentage:.1f}%")

Payment_Method = data["Payment Method"].value_counts()

for region, count in Payment_Method.items():
    percentage = (count / Payment_Method.sum()) * 100
    print(f"{region}: {percentage:.1f}%")

df_grouped = df.groupby(["month"])["Units Sold"].sum().reset_index()
df_grouped.set_index(["month"], inplace=True)
df_grouped.plot(kind="bar", color=["#FFD7C4"])

df_grouped = df.groupby(["day_name"])["Units Sold"].sum().reset_index()
df_grouped.set_index(["day_name"], inplace=True)
df_grouped.plot(kind="bar", color="#F1DEC6")

df_grouped = (
    df.groupby(["day_name", "Payment Method"])["Units Sold"].sum().reset_index()
)
df_grouped.set_index(["day_name", "Payment Method"], inplace=True)
df_grouped = (
    df.groupby(["day_name", "Payment Method"])["Units Sold"].sum().reset_index()
)

plt.figure(figsize=(12, 6))
sns.barplot(
    data=df_grouped,
    x="day_name",
    y="Units Sold",
    hue="Payment Method",
    palette=["#F1DEC6", "#FFD7C4", "#FFDBB5"],
    edgecolor="black",
)

plt.title("Units Sold by Payment Method and Day of the Week")
plt.xlabel(" day name")
plt.ylabel("Units Sold")
plt.legend(title="Payment Method")
plt.tight_layout()
plt.show()

df_grouped = df.groupby(["month", "Payment Method"])["Units Sold"].sum().reset_index()
df_grouped.set_index(["month", "Payment Method"], inplace=True)
df_grouped = df.groupby(["month", "Payment Method"])["Units Sold"].sum().reset_index()

plt.figure(figsize=(12, 6))
sns.barplot(
    data=df_grouped,
    x="month",
    y="Units Sold",
    hue="Payment Method",
    palette=["#F1DEC6", "#FFD7C4", "#FFDBB5"],
    edgecolor="black",
)

plt.title("Units Sold by Payment Method and month")
plt.xlabel("month")
plt.ylabel("Units Sold")
plt.legend(title="Payment Method")
plt.tight_layout()
plt.show()

data["Unit Price"].describe()
data["Total Revenue"].describe()

plt.figure(figsize=(5, 4))
sns.scatterplot(x="Unit Price", y="Total Revenue", data=df)

plt.title("Scatter Plot of Total Revenue vs Unit Price")
plt.xlabel("Unit Price")
plt.ylabel("Total Revenue")
plt.grid(True)
plt.show()

df.groupby(["Product Category"])["Total Revenue"].sum().reset_index()
df_grouped = (
    df.groupby(["Product Category", "Region"])["Total Revenue"].sum().reset_index()
)
df_grouped = (
    df.groupby(["Region", "Product Category"])["Total Revenue"].sum().reset_index()
)
df_grouped.set_index(["Region", "Product Category"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "Product Category"])["Total Revenue"]
    .sum()
    .reset_index()
)
df_grouped.set_index(["Payment Method", "Product Category"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "Region"])["Total Revenue"].sum().reset_index()
)
df_grouped.set_index(["Payment Method", "Region"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "Region", "Product Category"])["Units Sold"]
    .sum()
    .reset_index()
)
df_grouped.set_index(["Payment Method", "Region"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "Region", "Product Category"])["Total Revenue"]
    .sum()
    .reset_index()
)
df_grouped.set_index(["Payment Method", "Region", "Product Category"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "day_name"])["Total Revenue"].sum().reset_index()
)
df_grouped.set_index(["Payment Method", "day_name"], inplace=True)
df_grouped = (
    df.groupby(["Payment Method", "month"])["Total Revenue"].sum().reset_index()
)
df_grouped.set_index(["Payment Method", "month"], inplace=True)
df["Product Category"].unique()

revenue_summary = (
    df.groupby("Product Category")["Total Revenue"].agg(["min", "max"]).reset_index()
)
revenue_summary = (
    df.groupby(["Payment Method", "Product Category"])["Total Revenue"]
    .agg(["min", "max"])
    .reset_index()
)
revenue_summary.set_index(["Payment Method", "Product Category"], inplace=True)
print(revenue_summary)



from typing import Any, Dict, Optional

from pydantic import BaseModel, PrivateAttr


class CacheHandler(BaseModel):
    _cache: Dict[str, Any] = PrivateAttr(default_factory=dict)

    def add(self, tool, input, output):
        self._cache[f"{tool}-{input}"] = output

    def read(self, tool, input) -> Optional[str]:
        return self._cache.get(f"{tool}-{input}")



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field, PrivateAttr


class Printer:
    def print(self, content: str, color: Optional[str] = None):
        if color == "purple":
            self._print_purple(content)
        elif color == "red":
            self._print_red(content)
        elif color == "bold_green":
            self._print_bold_green(content)
        elif color == "bold_purple":
            self._print_bold_purple(content)
        elif color == "bold_blue":
            self._print_bold_blue(content)
        elif color == "yellow":
            self._print_yellow(content)
        elif color == "bold_yellow":
            self._print_bold_yellow(content)
        else:
            print(content)

    def _print_bold_purple(self, content):
        print("\033[1m\033[95m {}\033[00m".format(content))

    def _print_bold_green(self, content):
        print("\033[1m\033[92m {}\033[00m".format(content))

    def _print_purple(self, content):
        print("\033[95m {}\033[00m".format(content))

    def _print_red(self, content):
        print("\033[91m {}\033[00m".format(content))

    def _print_bold_blue(self, content):
        print("\033[1m\033[94m {}\033[00m".format(content))

    def _print_yellow(self, content):
        print("\033[93m {}\033[00m".format(content))

    def _print_bold_yellow(self, content):
        print("\033[1m\033[93m {}\033[00m".format(content))


class Logger(BaseModel):
    verbose: bool = Field(default=False)
    _printer: Printer = PrivateAttr(default_factory=Printer)

    def log(self, level, message, color="bold_yellow"):
        if self.verbose:
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self._printer.print(
                f"\n[{timestamp}][{level.upper()}]: {message}", color=color
            )



import json
import threading
import uuid
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Set, Optional, Tuple, Callable, Union, Type
from typing_extensions import Annotated

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator, create_model
from pydantic_core import PydanticCustomError

from versionhq._utils.process_config import process_config
from versionhq.task import TaskOutputFormat
from versionhq.tool.model import Tool, ToolCalled


class ResponseField(BaseModel):
    """
    Field class to use in the response schema for the JSON response.
    """

    title: str = Field(default=None)
    type: Type = Field(default=str)
    required: bool = Field(default=True)


    def _annotate(self, value: Any) -> Annotated:
        """
        Address `create_model`
        """
        return Annotated[self.type, value] if isinstance(value, self.type) else Annotated[str, str(value)]


    def _convert(self, value: Any) -> Any:
        try:
            if self.type is Any:
                pass
            elif self.type is int:
                return int(value)
            elif self.type is float:
                return float(value)
            elif self.type is list or self.type is dict:
                return json.loads(value)
            else:
                return value
        except:
            return value


    def create_pydantic_model(self, result: Dict, base_model: Union[BaseModel | Any]) -> Any:
        for k, v in result.items():
            if k is not self.title:
                pass
            elif type(v) is not self.type:
                v = self._convert(v)
                setattr(base_model, k, v)
            else:
                setattr(base_model, k, v)
        return base_model


class AgentOutput(BaseModel):
    """
    Keep adding agents' learning and recommendation and store it in `pydantic` field of `TaskOutput` class.
    Since the TaskOutput class has `agent` field, we don't add any info on the agent that handled the task.
    """
    customer_id: str = Field(default=None, max_length=126, description="customer uuid")
    customer_analysis: str = Field(default=None, max_length=256, description="analysis of the customer")
    product_overview: str = Field(default=None, max_length=256, description="analysis of the client's business")
    usp: str = Field()
    cohort_timeframe: int = Field(default=None, max_length=256, description="suitable cohort timeframe in days")
    kpi_metrics: List[str] = Field(default=list, description="Ideal KPIs to be tracked")
    assumptions: List[Dict[str, Any]] = Field(default=list, description="assumptions to test")



class TaskOutput(BaseModel):
    """
    Store the final output of the task in TaskOutput class.
    Depending on the task output format, use `raw`, `pydantic`, `json_dict` accordingly.
    """

    task_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store Task ID")
    raw: str = Field(default="", description="Raw output of the task")
    json_dict: Union[Dict[str, Any]] = Field(default=None, description="`raw` converted to dictionary")
    pydantic: Optional[Any] = Field(default=None, description="`raw` converted to the abs. pydantic model")

    def __str__(self) -> str:
        return str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw

    @property
    def json(self) -> Optional[str]:
        if self.output_format != TaskOutputFormat.JSON:
            raise ValueError(
                """
                Invalid output format requested.
                If you would like to access the JSON output,
                pleae make sure to set the output_json property for the task
                """
            )
        return json.dumps(self.json_dict)


    def to_dict(self) -> Dict[str, Any]:
        """
        Convert pydantic / raw output into dict and return the dict.
        When we only have `raw` output, return `{ output: raw }` to avoid an error
        """

        output_dict = {}
        if self.json_dict:
            output_dict.update(self.json_dict)
        elif self.pydantic:
            output_dict.update(self.pydantic.model_dump())
        else:
            output_dict.upate({ "output": self.raw })
        return output_dict


    def context_prompting(self) -> str:
        """
        When the task is called as context, return its output in concise string to add it to the prompt
        """
        return json.dumps(self.json_dict) if self.json_dict else self.raw[0: 127]



class Task(BaseModel):
    """
    Task to be executed by the agent or the team.
    Each task must have a description and at least one expected output format either Pydantic, Raw, or JSON, with necessary fields in ResponseField.
    Then output will be stored in TaskOutput class.
    """

    __hash__ = object.__hash__

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="unique identifier for the object, not set by user")
    name: Optional[str] = Field(default=None)
    description: str = Field(description="Description of the actual task")
    _original_description: str = PrivateAttr(default=None)

    # output
    expected_output_json: bool = Field(default=True)
    expected_output_pydantic: bool = Field(default=False)
    output_field_list: List[ResponseField] = Field(
        default=[ResponseField(title="output", type=str, required=False)],
        description="provide output key and data type. this will be cascaded to the agent via task.prompt()"
    )
    output: Optional[TaskOutput] = Field(default=None, description="store the final task output in TaskOutput class")

    # task setup
    context: Optional[List["Task"]] = Field(default=None, description="other tasks whose outputs should be used as context")
    tools_called: Optional[List[ToolCalled]] = Field(default_factory=list, description="tools that the agent can use for this task")
    take_tool_res_as_final: bool = Field(default=False, description="when set True, tools res will be stored in the `TaskOutput`")
    allow_delegation: bool = Field(default=False, description="ask other agents for help and run the task instead")

    prompt_context: Optional[str] = Field(default=None)
    async_execution: bool = Field(default=False,description="whether the task should be executed asynchronously or not")
    config: Optional[Dict[str, Any]] = Field(default=None, description="configuration for the agent")
    callback: Optional[Any] = Field(default=None, description="callback to be executed after the task is completed.")

    # recording
    processed_by_agents: Set[str] = Field(default_factory=set)
    used_tools: int = 0
    tools_errors: int = 0
    delegations: int = 0


    @property
    def output_prompt(self) -> str:
        """
        Draft prompts on the output format by converting `output_field_list` to dictionary.
        """

        output_prompt, output_formats_to_follow = "", dict()
        for item in self.output_field_list:
            output_formats_to_follow[item.title] = f"<Return your answer in {item.type.__name__}>"

        output_prompt = f"""
Your outputs MUST adhere to the following format and should NOT include any irrelevant elements:
{output_formats_to_follow}
        """
        return output_prompt


    @property
    def expected_output_formats(self) -> List[TaskOutputFormat]:
        """
        Return output formats in list with the ENUM item.
        `TaskOutputFormat.RAW` is set as default.
        """
        outputs = [TaskOutputFormat.RAW,]
        if self.expected_output_json:
            outputs.append(TaskOutputFormat.JSON)
        if self.expected_output_pydantic:
            outputs.append(TaskOutputFormat.PYDANTIC)
        return outputs


    @property
    def key(self) -> str:
        output_format = (
            TaskOutputFormat.JSON
            if self.expected_output_json == True
            else (
                TaskOutputFormat.PYDANTIC
                if self.expected_output_pydantic == True
                else TaskOutputFormat.RAW
            )
        )
        source = [self.description, output_format]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def summary(self) -> str:
        return f"""
        Task: {self.id} - {self.description}
        "task_description": {self.description}
        "task_expected_output": {self.output_prompt}
        "task_tools": {", ".join([tool_called.tool.name for tool_called in self.tools_called])}
        """


    # validators
    @model_validator(mode="before")
    @classmethod
    def process_model_config(cls, values: Dict[str, Any]):
        return process_config(values_to_update=values, model_class=cls)


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_required_fields(self):
        required_fields = ["description",]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError( f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_attributes_based_on_config(self) -> "Task":
        """
        Set attributes based on the agent configuration.
        """

        if self.config:
            for key, value in self.config.items():
                setattr(self, key, value)
        return self


    ## comment out as we set raw as the default TaskOutputFormat
    # @model_validator(mode="after")
    # def validate_output_format(self):
    #     if self.expected_output_json == False  and self.expected_output_pydantic == False:
    #         raise PydanticCustomError("Need to choose at least one output format.")
    #     return self


    @model_validator(mode="after")
    def backup_description(self):
        if self._original_description == None:
            self._original_description = self.description
        return self


    def prompt(self, customer: str = None, product_overview: str = None) -> str:
        """
        Format the task prompt and cascade it to the agent.
        When the task has context, add context prompting of all the tasks in the context.
        When we have cusotmer/product info, add them to the prompt.
        """

        task_slices = [self.description, f"{self.output_prompt}", f"Take the following context into consideration: "]

        if self.context:
            context_outputs = "\n".join([task.output.context_prompting() if hasattr(task, "output") else "" for task in self.context])
            task_slices.insert(len(task_slices),  context_outputs)

        if customer:
            task_slices.insert(len(task_slices),  f"Customer overview: {customer}")

        if product_overview:
            task_slices.insert(len(task_slices), f"Product overview: {product_overview}")

        if self.prompt_context:
            task_slices.insert(len(task_slices), self.prompt_context)

        return "\n".join(task_slices)


    def create_json_output(self, raw_result: Any) -> Optional[Dict[str, Any]]:
        """
        Create json (dict) output from the raw result.
        """

        output_json_dict: Optional[Dict[str, Any]] = None

        if isinstance(raw_result, BaseModel):
            output_json_dict = raw_result.model_dump()

        elif isinstance(raw_result, dict):
            output_json_dict = raw_result

        elif isinstance(raw_result, str):
            try:
                output_json_dict = json.loads(raw_result)
            except json.JSONDecodeError:
                try:
                    output_json_dict = eval(raw_result)
                except:
                    try:
                        import ast
                        output_json_dict = ast.literal_eval(raw_result)
                    except:
                        output_json_dict = { "output": raw_result }

        return output_json_dict



    def create_pydantic_output(self, output_json_dict: Dict[str, Any], raw_result: Any = None) -> Optional[Any]:
        """
        Create pydantic output from the `raw` result.
        """

        output_pydantic = None
        if isinstance(raw_result, BaseModel):
            output_pydantic = raw_result

        elif hasattr(output_json_dict, "output"):
            output_pydantic = create_model("PydanticTaskOutput", output=output_json_dict["output"], __base__=BaseModel)

        else:
            output_pydantic = create_model("PydanticTaskOutput", __base__=BaseModel)
            try:
                for item in self.output_field_list:
                    value = output_json_dict[item.title] if hasattr(output_json_dict, item.title) else None
                    if value and type(value) is not item.type:
                        value = item._convert(value)
                    setattr(output_pydantic, item.title, value)
            except:
                setattr(output_pydantic, "output", output_json_dict)

        return output_pydantic


    def _get_output_format(self) -> TaskOutputFormat:
        if self.output_json == True:
            return TaskOutputFormat.JSON
        if self.output_pydantic == True:
            return TaskOutputFormat.PYDANTIC
        return TaskOutputFormat.RAW


    def interpolate_inputs(self, inputs: Dict[str, Any]) -> None:
        """
        Interpolate inputs into the task description and expected output.
        """
        if inputs:
            self.description = self._original_description.format(**inputs)
            # self.expected_output = self._original_expected_output.format(**inputs)


    # task execution
    def execute_sync(self, agent, context: Optional[str] = None, callback_kwargs: Dict[str, Any] = None) -> TaskOutput:
        """
        Execute the task synchronously.
        When the task has context, make sure we have executed all the tasks in the context first.
        """

        if self.context:
            for task in self.context:
                if task.output is None:
                    task._execute_core(agent, context, callback_kwargs)

        return self._execute_core(agent, context)


    def execute_async(self, agent, context: Optional[str] = None, callback_kwargs: Dict[str, Any] = None) -> Future[TaskOutput]:
        """
        Execute the task asynchronously.
        """

        future: Future[TaskOutput] = Future()
        threading.Thread(
            daemon=True,
            target=self._execute_task_async,
            args=(agent, context, callback_kwargs, future),
        ).start()
        return future


    def _execute_task_async(self, agent, context: Optional[str], callback_kwargs: Dict[str, Any], future: Future[TaskOutput]) -> None:
        """
        Execute the task asynchronously with context handling.
        """

        result = self._execute_core(agent, context, callback_kwargs)
        future.set_result(result)


    def _execute_core(self, agent, context: Optional[str], callback_kwargs: Optional[Dict[str, Any]] = None) -> TaskOutput:
        """
        Run the core execution logic of the task.
        To speed up the process, when the format is not expected to return, we will skip the conversion process.
        """
        from versionhq.agent.model import Agent

        self.prompt_context = context

        if self.allow_delegation:
            agent = Agent(role="delegated_agent", goal=agent.goal, llm=agent.llm) #! REFINEME - logic to pick up the high performer
            self.delegations += 1

        output_raw = agent.execute_task(task=self, context=context)
        output_json_dict = self.create_json_output(raw_result=output_raw) if self.expected_output_json is True else None
        output_pydantic = self.create_pydantic_output(output_json_dict=output_json_dict) if self.expected_output_pydantic else None
        task_output = TaskOutput(
            task_id=self.id,
            raw=output_raw,
            pydantic=output_pydantic,
            json_dict=output_json_dict
        )
        self.output = task_output
        self.processed_by_agents.add(agent.role)

        # self._set_end_execution_time(start_time)

        if self.callback:
            if isinstance(self.callback, Callable):
                self.callback(**callback_kwargs)
            else:
                self.callback(self.output)

        # if self._execution_span:
        #     # self._telemetry.task_ended(self._execution_span, self, agent.team)
        #     self._execution_span = None

        # if self.output_file:
        #     content = (
        #         json_output
        #         if json_output
        #         else pydantic_output.model_dump_json() if pydantic_output else result
        #     )
        #     self._save_file(content)
        return task_output


class ConditionalTask(Task):
    """
    A task that can be conditionally executed based on the output of another task.
    Use this with `Team`.
    """

    condition: Callable[[TaskOutput], bool] = Field(
        default=None,
        description="max. number of retries for an agent to execute a task when an error occurs.",
    )

    def __init__(
        self,
        condition: Callable[[Any], bool],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.condition = condition

    def should_execute(self, context: TaskOutput) -> bool:
        """
        Decide whether the conditional task should be executed based on the provided context.
        Return `True` if it should be executed.
        """
        return self.condition(context)

    def get_skipped_task_output(self):
        return TaskOutput(task_id=self.id, raw="", pydantic=None, json_dict=None)



from typing import Any, Optional, Union
from versionhq.tool.model import ToolCalled, InstructorToolCalled, CacheTool
from versionhq._utils.cache_handler import CacheHandler


class ToolHandler:
    """
    Record the tool usage by ToolCalled instance with cache and error recording.
    Use as a callback function.
    """

    last_used_tool: ToolCalled = {}
    cache: Optional[CacheHandler]
    error: Optional[str]

    def __init__(self, cache_handler: Optional[CacheHandler] = None):
        """
        Initialize the callback handler.
        """

        self.cache = cache_handler
        self.last_used_tool = {}

    def record_last_tool_used(
        self,
        last_used_tool: Union[ToolCalled, InstructorToolCalled],
        output: str,
        should_cache: bool = True,
    ) -> Any:
        self.last_used_tool = last_used_tool

        if self.cache and should_cache and last_used_tool.tool_name != CacheTool().name:
            self.cache.add(
                last_used_tool.tool_name,
                input.last_used_tool.arguments,
                output=output,
            )

    def has_called_before(self, tool_called: ToolCalled = None) -> bool:
        if tool_called is None or not self.last_used_tool:
            return False

        if tool_called := self.last_used_tool:
            return bool(
                (tool_called.tool.name == self.last_used_tool.tool.name)
                and (tool_called.arguments == self.last_used_tool.arguments)
            )



from pydantic import BaseModel, Field


class UsageMetrics(BaseModel):
    """
    Model to track usage metrics for the agent/team's execution.
    """

    total_tokens: int = Field(default=0, description="Total number of tokens used")
    prompt_tokens: int = Field(default=0, description="Number of tokens used in prompts")
    cached_prompt_tokens: int = Field(default=0, description="Number of cached prompt tokens used")
    completion_tokens: int = Field(default=0, description="Number of tokens used in completions")
    successful_requests: int = Field(default=0, description="Number of successful requests made")

    def add_usage_metrics(self, usage_metrics: "UsageMetrics"):
        """
        Add the usage metrics from another UsageMetrics object.
        """
        self.total_tokens += usage_metrics.total_tokens
        self.prompt_tokens += usage_metrics.prompt_tokens
        self.cached_prompt_tokens += usage_metrics.cached_prompt_tokens
        self.completion_tokens += usage_metrics.completion_tokens
        self.successful_requests += usage_metrics.successful_requests



BACKSTORY="""You are a {role} with deep understanding of {knowledge} and highly skilled in {skillsets}.
You have access to call the RAG tools that can {rag_tool_overview}. By leveraging these tools, your knowledge, and skillsets, you can identify competitive strategies that have been proven effective to achieve the goal: {goal}. Take these into consideration, create innovative solutions.
"""



import warnings

warnings.filterwarnings(
    "ignore",
    message="Pydantic serializer warnings:",
    category=UserWarning,
    module="pydantic.main",
)

from versionhq.agent.model import Agent
from versionhq.clients.customer.model import Customer
from versionhq.clients.product.model import Product, ProductProvider
from versionhq.clients.workflow.model import MessagingWorkflow, MessagingComponent, Score, ScoreFormat
from versionhq.llm.model import LLM
from versionhq.task.model import Task, TaskOutput
from versionhq.team.model import Team, TeamOutput
from versionhq.tool.model import Tool


__version__ = "1.1.7.3"
__all__ = [
    "Agent",
    "Customer",
    "Product",
    "ProductProvider",
    "MessagingWorkflow",
    "MessagingComponent",
    "Score",
    "ScoreFormat",
    "LLM",
    "Task",
    "TaskOutput",
    "Team",
    "TeamOutput",
    "Tool",
]



import os
import uuid
from abc import ABC
from typing import Any, Dict, List, Optional, TypeVar, Union
from dotenv import load_dotenv
from pydantic import UUID4, BaseModel, Field, InstanceOf, PrivateAttr, model_validator, field_validator
from pydantic_core import PydanticCustomError

from versionhq._utils.cache_handler import CacheHandler
from versionhq._utils.logger import Logger
from versionhq._utils.rpm_controller import RPMController
from versionhq._utils.usage_metrics import UsageMetrics
from versionhq.agent.parser import AgentAction
from versionhq.llm.llm_vars import LLM_VARS
from versionhq.llm.model import LLM, DEFAULT_CONTEXT_WINDOW
from versionhq.task import TaskOutputFormat
from versionhq.task.model import ResponseField
from versionhq.tool.model import Tool, ToolCalled
from versionhq.tool.tool_handler import ToolHandler

load_dotenv(override=True)
T = TypeVar("T", bound="Agent")


# def _format_answer(agent, answer: str) -> Union[AgentAction, AgentFinish]:
#     return AgentParser(agent=agent).parse(answer)

# def mock_agent_ops_provider():
#     def track_agent(*args, **kwargs):
#         def noop(f):
#             return f
#         return noop
#     return track_agent

# track_agent = mock_agent_ops_provider()

# agentops = None
# if os.environ.get("AGENTOPS_API_KEY"):
#     try:
#         from agentops import track_agent
#     except ImportError:
#         track_agent = mock_agent_ops_provider()
# else:
#    track_agent = mock_agent_ops_provider()


class TokenProcess:
    total_tokens: int = 0
    prompt_tokens: int = 0
    cached_prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0

    def sum_prompt_tokens(self, tokens: int):
        self.prompt_tokens = self.prompt_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_completion_tokens(self, tokens: int):
        self.completion_tokens = self.completion_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_cached_prompt_tokens(self, tokens: int):
        self.cached_prompt_tokens = self.cached_prompt_tokens + tokens

    def sum_successful_requests(self, requests: int):
        self.successful_requests = self.successful_requests + requests

    def get_summary(self) -> UsageMetrics:
        return UsageMetrics(
            total_tokens=self.total_tokens,
            prompt_tokens=self.prompt_tokens,
            cached_prompt_tokens=self.cached_prompt_tokens,
            completion_tokens=self.completion_tokens,
            successful_requests=self.successful_requests,
        )


# @track_agent()
class Agent(ABC, BaseModel):
    """
    Agent class that run on LLM.
    Agents execute tasks alone or in the team, using RAG tools and knowledge base if any.
    Agents will prioritize team tasks when they belong to the team.
    * (Temp) Comment out all the optional fields except for Team and LLM settings for convenience.
    """

    __hash__ = object.__hash__
    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=False))
    _rpm_controller: Optional[RPMController] = PrivateAttr(default=None)
    _request_within_rpm_limit: Any = PrivateAttr(default=None)
    _token_process: TokenProcess = PrivateAttr(default_factory=TokenProcess)
    _times_executed: int = PrivateAttr(default=0)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    role: str = Field(description="role of the agent - used in summary and logs")
    goal: str = Field(description="concise goal of the agent (details are set in the Task instance)")
    backstory: Optional[str] = Field(default=None, description="system context passed to the LLM")
    knowledge: Optional[str] = Field(default=None, description="external knowledge fed to the agent")
    skillsets: Optional[List[str]] = Field(default_factory=list)

    # tools
    tools: Optional[List[Any]] = Field(default_factory=list)
    tool_handler: InstanceOf[ToolHandler] = Field(default=None, description="handle tool cache and last used tool")

    # team, task execution rules
    team: Optional[List[Any]] = Field(default=None, description="Team to which the agent belongs")
    allow_delegation: bool = Field(default=False,description="Enable agent to delegate and ask questions among each other")
    allow_code_execution: Optional[bool] = Field(default=False, description="Enable code execution for the agent.")
    max_retry_limit: int = Field(default=2,description="max. number of retries for the task execution when an error occurs. cascaed to the `invoke` function")
    max_iter: Optional[int] = Field(default=25,description="max. number of iterations for an agent to execute a task")
    step_callback: Optional[Any] = Field(default=None,description="Callback to be executed after each step of the agent execution")

    # llm settings cascaded to the LLM model
    llm: Union[str, InstanceOf[LLM], Any] = Field(default=None)
    function_calling_llm: Union[str, InstanceOf[LLM], Any] = Field(default=None)
    respect_context_window: bool = Field(default=True,description="Keep messages under the context window size by summarizing content")
    max_tokens: Optional[int] = Field(default=None, description="max. number of tokens for the agent's execution")
    max_execution_time: Optional[int] = Field(default=None, description="max. execution time for an agent to execute a task")
    max_rpm: Optional[int] = Field(default=None, description="max. number of requests per minute for the agent execution")

    # prompt rules
    use_system_prompt: Optional[bool] = Field(default=True, description="Use system prompt for the agent")
    system_template: Optional[str] = Field(default=None, description="System format for the agent.")
    prompt_template: Optional[str] = Field(default=None, description="Prompt format for the agent.")
    response_template: Optional[str] = Field(default=None, description="Response format for the agent.")

    # config, cache, error handling
    config: Optional[Dict[str, Any]] = Field(default=None, exclude=True, description="Configuration for the agent")
    cache: bool = Field(default=True, description="Whether the agent should use a cache for tool usage.")
    cache_handler: InstanceOf[CacheHandler] = Field(default=None, description="An instance of the CacheHandler class.")
    formatting_errors: int = Field(default=0, description="Number of formatting errors.")
    verbose: bool = Field(default=True, description="Verbose mode for the Agent Execution")
    agent_ops_agent_name: str = None
    agent_ops_agent_id: str = None


    def __repr__(self):
        return f"Agent(role={self.role}, goal={self.goal}, backstory={self.backstory})"


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_required_fields(self):
        required_fields = ["role", "goal"]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError( f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_llm(self):
        """
        Set up the base model and function calling model (if any) using the LLM class.
        Pass the model config params: `llm`, `max_tokens`, `max_execution_time`, `step_callback`,`respect_context_window` to the LLM class.
        The base model is selected on the client app, else use the default model.
        """

        self.agent_ops_agent_name = self.role
        unaccepted_attributes = ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_REGION_NAME"]
        callbacks = ([self.step_callback,]if self.step_callback is not None else [])

        if isinstance(self.llm, LLM):
            self.llm.timeout = self.max_execution_time
            self.llm.max_tokens = self.max_tokens
            self.llm.context_window_size = (self.llm.get_context_window_size() if self.respect_context_window == True else DEFAULT_CONTEXT_WINDOW)
            self.llm.callbacks = callbacks

        elif isinstance(self.llm, str) or self.llm is None:
            model_name = os.environ.get("LITELLM_MODEL_NAME", os.environ.get("MODEL", "gpt-3.5-turbo"))
            llm_params = {
                "model": model_name if self.llm is None else self.llm,
                "timeout": self.max_execution_time,
                "max_tokens": self.max_tokens,
                "callbacks": callbacks,
                "api_key": os.environ.get("LITELLM_API_KEY", None),
                "base_url": os.environ.get("OPENAI_API_BASE", os.environ.get("OPENAI_BASE_URL", None))
            }

            set_provider = model_name.split("/")[0] if "/" in model_name else "openai" #! REFINEME
            for provider, env_vars in LLM_VARS.items():
                if provider == set_provider:
                    for env_var in env_vars:
                        key_name = env_var.get("key_name")

                        if key_name and key_name not in unaccepted_attributes:
                            env_value = os.environ.get(key_name)
                            if env_value:
                                key_name = ("api_key" if "API_KEY" in key_name else key_name)
                                key_name = ("api_base" if "API_BASE" in key_name else key_name)
                                key_name = ("api_version" if "API_VERSION" in key_name else key_name)
                                llm_params[key_name] = env_value
                        elif env_var.get("default", False):
                            for key, value in env_var.items():
                                if key not in ["prompt", "key_name", "default"]:
                                    if key in os.environ:
                                        llm_params[key] = value
            self.llm = LLM(**llm_params)
            context_window_size = (self.llm.get_context_window_size() if self.respect_context_window == True else DEFAULT_CONTEXT_WINDOW)
            self.llm.context_window_size = context_window_size

        else:
            llm_params = {
                "model": (getattr(self.llm, "model_name") or getattr(self.llm, "deployment_name") or str(self.llm)),
                "max_tokens": (getattr(self.llm, "max_tokens") or self.max_tokens or 3000),
                "timeout": getattr(self.llm, "timeout", self.max_execution_time),
                "callbacks": getattr(self.llm, "callbacks") or callbacks,
                "temperature": getattr(self.llm, "temperature", None),
                "logprobs": getattr(self.llm, "logprobs", None),
                "api_key": getattr(self.llm, "api_key", os.environ.get("LITELLM_API_KEY", None)),
                "base_url": getattr(self.llm, "base_url", None),
                "organization": getattr(self.llm, "organization", None),
            }
            llm_params = {  k: v for k, v in llm_params.items() if v is not None }
            self.llm = LLM(**llm_params)

        """
        Set up funcion_calling LLM as well. For the sake of convenience, use the same metrics as the base LLM settings.
        """
        if self.function_calling_llm:
            if isinstance(self.function_calling_llm, LLM):
                self.function_calling_llm.timeout = self.max_execution_time
                self.function_calling_llm.max_tokens = self.max_tokens
                self.function_calling_llm.callbacks = callbacks
                context_window_size = (
                    self.function_calling_llm.get_context_window_size()
                    if self.respect_context_window == True
                    else DEFAULT_CONTEXT_WINDOW
                )
                self.function_calling_llm.context_window_size = context_window_size

            elif isinstance(self.function_calling_llm, str):
                self.function_calling_llm = LLM(
                    model=self.function_calling_llm,
                    timeout=self.max_execution_time,
                    max_tokens=self.max_tokens,
                    callbacks=callbacks,
                )
                context_window_size = (
                    self.function_calling_llm.get_context_window_size()
                    if self.respect_context_window == True
                    else DEFAULT_CONTEXT_WINDOW
                )
                self.function_calling_llm.context_window_size = context_window_size

            else:
                model_name = getattr(
                    self.function_calling_llm,
                    "model_name",
                    getattr(
                        self.function_calling_llm,
                        "deployment_name",
                        str(self.function_calling_llm),
                    ),
                )
                if model_name is not None or model_name != "":
                    self.function_calling_llm = LLM(
                        model=model_name,
                        timeout=self.max_execution_time,
                        max_tokens=self.max_tokens,
                        callbacks=callbacks,
                    )
        return self


    @model_validator(mode="after")
    def set_up_tools(self):
        """
        Similar to the LLM set up, when the agent has tools, we will declare them using the Tool class.
        """

        if not self.tools:
            pass

        else:
            tools_in_class_format = []
            for tool in self.tools:
                if isinstance(tool, Tool):
                    tools_in_class_format.append(tool)
                elif isinstance(tool, str):
                    tool_to_add = Tool(name=tool)
                    tools_in_class_format.append(tool_to_add)
                else:
                    pass
            self.tools = tools_in_class_format

        return self


    @model_validator(mode="after")
    def set_up_backstory(self):
        """
        Set up the backstory using a templated BACKSTORY when the backstory is None
        """

        if self.backstory is None:
            from versionhq.agent.TEMPLATES.Backstory import BACKSTORY
            backstory = BACKSTORY.format(
                role=self.role,
                knowledge=self.knowledge if isinstance(self.knowledge, str) else None,
                skillsets=", ".join([item for item in self.skillsets]),
                rag_tool_overview=", ".join([item.name for item in self.tools]),
                goal=self.goal,
            )
            self.backstory = backstory

        return self


    def invoke(
            self, prompts: str, output_formats: List[TaskOutputFormat], response_fields: List[ResponseField], **kwargs
            ) -> Dict[str, Any]:
        """
        Receive the system prompt in string and create formatted prompts using the system prompt and the agent's backstory.
        Then call the base model.
        When encountering errors, we try the task execution up to `self.max_retry_limit` times.
        """

        task_execution_counter = 0

        messages = []
        messages.append({"role": "user", "content": prompts})  #! REFINEME
        messages.append({"role": "assistant", "content": self.backstory})
        print("Messages sent to the model:", messages)

        callbacks = kwargs.get("callbacks", None)

        response = self.llm.call(
            messages=messages,
            output_formats=output_formats,
            field_list=response_fields,
            callbacks=callbacks,
        )
        task_execution_counter += 1
        print("Agent's #1 res: ", response)

        if (response is None or response == "") and task_execution_counter < self.max_retry_limit:
            while task_execution_counter <= self.max_retry_limit:
                response = self.llm.call(
                    messages=messages,
                    output_formats=output_formats,
                    field_list=response_fields,
                    callbacks=callbacks,
                )
                task_execution_counter += 1
                print(f"Agent's #{task_execution_counter} res: ", response)

        elif response is None or response == "":
            print("Received None or empty response from LLM call.")
            raise ValueError("Invalid response from LLM call - None or empty.")

        return {"output": response.output if hasattr(response, "output") else response}


    def execute_task(self, task, context: Optional[str] = None, tools: Optional[str] = None) -> str:
        """
        Execute the task and return the output in string.
        To simplify, the tools are cascaded from the `tools_called` under the `task` Task instance if any.
        When the tools are given, the agent must use them.
        The agent must consider the context to excute the task as well when it is given.
        """

        task_prompt = task.prompt()
        if context is not task.prompt_context:  # as `task.prompt()` includes adding `task.prompt_context` to the prompt.
            task_prompt += context

        tool_results = []
        if task.tools_called:
            for tool_called in task.tools_called:
                tool_result = tool_called.tool.run()
                tool_results.append(tool_result)

            if task.take_tool_res_as_final:
                return tool_results

        # if self.team and self.team._train:
        #     task_prompt = self._training_handler(task_prompt=task_prompt)
        # else:
        #     task_prompt = self._use_trained_data(task_prompt=task_prompt)

        try:
            result = self.invoke(
                prompts=task_prompt,
                output_formats=task.expected_output_formats,
                response_fields=task.output_field_list,
            )["output"]

        except Exception as e:
            self._times_executed += 1
            if self._times_executed > self.max_retry_limit:
                raise e
            result = self.execute_task(
                task, context, [tool_called.tool for tool_called in task.tools_called]
            )

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller.stop_rpm_counter()

        # for tool_result in self.tools_results:
        #     if tool_result.get("result_as_answer", False):
        #         result = tool_result["result"]

        return result



import uuid
from abc import ABC
from datetime import date, datetime, time, timedelta
from typing import Any, Dict, List, Union, Callable, Type, Optional, get_args, get_origin
from pydantic import UUID4, InstanceOf, BaseModel, ConfigDict, Field, create_model, field_validator, model_validator
from pydantic_core import PydanticCustomError

from versionhq.clients.product.model import Product
from versionhq.clients.customer.model import Customer
from versionhq.agent.model import Agent
from versionhq.team.model import Team


class ScoreFormat:
    def __init__(self, rate: Union[float, int] = 0, weight: int = 1):
        self.rate = rate
        self.weight = weight
        self.aggregate = rate * weight


class Score:
    """
    Evaluate the score on 0 (no performance) to 1 scale.
    `rate`: Any float from 0.0 to 1.0 given by an agent.
    `weight`: Importance of each factor to the aggregated score.
    """

    def __init__(
        self,
        brand_tone: ScoreFormat = ScoreFormat(0, 0),
        audience: ScoreFormat = ScoreFormat(0, 0),
        track_record: ScoreFormat = ScoreFormat(0, 0),
        **kwargs: Optional[Dict[str, ScoreFormat]],
    ):
        self.brand_tone = brand_tone
        self.audience = audience
        self.track_record = track_record
        self.kwargs = kwargs


    def result(self) -> int:
        aggregate_score = self.brand_tone.aggregate + self.audience.aggregate + self.track_record.aggregate
        denominator = self.brand_tone.weight + self.audience.weight + self.track_record.weight

        for k, v in self.kwargs.items():
            aggregate_score += v.aggregate
            denominator += v.weight

        if denominator == 0:
            return 0

        return round(aggregate_score / denominator, 2)



class MessagingComponent(ABC, BaseModel):
    layer_id: int = Field(default=0, description="add id of the layer: 0, 1, 2")
    message: str = Field(default=None, max_length=1024, description="text message content to be sent")
    interval: Optional[str] = Field(
        default=None,description="interval to move on to the next layer. if this is the last layer, set as `None`")
    score: Union[float, InstanceOf[Score]] = Field(default=None)


    def store_scoring_result(self, scoring_subject: str, score: Union[int, Score, ScoreFormat] = None):
        """
        Set up the `score` field
        """

        if isinstance(score, Score):
            setattr(self, "score", score)

        elif isinstance(score, ScoreFormat):
            score_instance = Score()
            setattr(score_instance, scoring_subject, score)
            setattr(self, "score", score_instance)

        elif isinstance(score, int) or isinstance(score, float):
            score_instance, score_format_instance = Score(), ScoreFormat(rate=score, weight=1)
            setattr(score_instance, "kwargs", { scoring_subject: score_format_instance })
            setattr(self, "score", score_instance)

        else:
            pass

        return self



class MessagingWorkflow(ABC, BaseModel):
    """
    Store 3 layers of messaging workflow sent to `customer` on the `product`
    """

    _created_at: Optional[datetime]
    _updated_at: Optional[datetime]

    model_config = ConfigDict()

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    components: List[MessagingComponent] = Field(default_factory=list, description="store messaging components in the workflow")

    # responsible tean or agents
    team: Optional[Team] = Field(default=None, description="store `Team` instance responsibile for autopiloting this workflow")
    agents: Optional[List[Agent]] = Field(
        default=None, description="store `Agent` instances responsible for autopiloting this workflow. if the team exsits, this field remains as `None`")

    # metrics
    destination: Optional[str] = Field( default=None, description="destination service to launch this workflow")
    product: InstanceOf[Product] = Field(default=None)
    customer: InstanceOf[Customer] = Field(default=None)

    metrics: Union[List[Dict[str, Any]], List[str]] = Field(
        default=None, max_length=256, description="store metrics that used to predict and track the performance of this workflow.")


    @property
    def name(self):
        if self.customer.id:
            return f"Workflow ID: {self.id} - on {self.product.id} for {self.customer.id}"
        else:
            return f"Workflow ID: {self.id} - on {self.product.id}"


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def set_up_destination(self):
        """
        Set up the destination service when self.destination is None.
        Prioritize customer's destination to the product provider's destination list.
        """
        if self.destination is None:
            if self.customer is not None:
                self.destination = self.customer.on

            else:
                destination_list = self.product.provider.destinations
                if destination_list:
                    self.destination = destination_list[0]
        return self


    def reassign_agent_or_team(self, agents: List[Agent] = None, team: Team = None) -> None:
        """
        Fire unresponsible agents/team and assign new one.
        """

        if not agents and not team:
            raise ValueError("Need to add at least 1 agent or team.")

        self.agents = agents
        self.team = team
        self.updated_at = datetime.datetime.now()



import os
from dotenv import load_dotenv
from typing import Any, List, Optional
from pydantic import BaseModel, Field

from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField

load_dotenv(override=True)


class TeamPlanner:
    """
    (Optional) Plan how the team should handle multiple tasks using LLM.
    """

    def __init__(self, tasks: List[Task], planner_llm: Optional[Any] = None):
        self.tasks = tasks
        self.planner_llm = (
            planner_llm if planner_llm != None else os.environ.get("LITELLM_MODEL_NAME")
        )

    def _handle_task_planning(self) -> BaseModel:
        """
        Handles the team planning by creating detailed step-by-step plans for each task.
        """

        planning_agent = Agent(
            role="Task Execution Planner",
            goal="Your goal is to create an extremely detailed, step-by-step plan based on the tasks and tools available to each agent so that they can perform the tasks in an exemplary manner",
            backstory="You have a strong ability to design efficient organizational structures and task processes, minimizing unnecessary steps.",
            llm=self.planner_llm,
        )

        task_summary_list = [task.summary for task in self.tasks]
        task_to_handle = Task(
            description=f"""
                Based on the following task summaries, create the most descriptive plan that the team can execute most efficiently. Take all the task summaries - task's description and tools available - into consideration. Your answer only contains a dictionary.

                Task summaries: {" ".join(task_summary_list)}
             """,
            expected_output_json=False,
            expected_output_pydantic=True,
            output_field_list=[
                ResponseField(title=f"{task.id}", type=str, required=True)
                for task in self.tasks
            ],
        )
        task_output = task_to_handle.execute_sync(agent=planning_agent)

        if isinstance(task_output.pydantic, BaseModel):
            return task_output.pydantic

        else:
            return None



import uuid
from typing import Any, Dict, List, Callable, Type, Optional, get_args, get_origin
from pydantic import UUID4, InstanceOf, BaseModel, ConfigDict, Field, create_model, field_validator, model_validator
from pydantic_core import PydanticCustomError


class ProductProvider(BaseModel):
    """
    Store the minimal client information.
    `data_pipeline` and `destinations` are for composio plug-in.
    (!REFINEME) Create an Enum list for the options.
    (!REFINEME) Create an Enum list for regions.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="client name")
    region: Optional[str] = Field(default=None, description="region of client's main business operation")
    data_pipeline: Optional[List[str]] = Field(default=None, description="store the data pipelines that the client is using")
    destinations: Optional[List[str]] = Field(default=None,description="store the destination services that the client is using")

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})



class Product(BaseModel):
    """
    Store the product information necessary to the outbound effrots and connect it to the `ProductProvider` instance.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="product name")
    description: Optional[str] = Field(
        default=None,max_length=256,description="product description scraped from landing url or client input. cascade to the agent")
    provider: Optional[ProductProvider] = Field(default=None)
    audience: Optional[str] = Field(default=None, description="target audience")
    usp: Optional[str] = Field(default=None)
    landing_url: Optional[str] = Field(default=None, description="marketing url of the product if any")
    cohort_timeframe: Optional[int] = Field(default=30, description="ideal cohort timeframe of the product in days")


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})
